# -*- coding: utf-8 -*-
"""This module contains a parser for the Peruvian CPI data."""

import re
from datetime import date

import pandas as pd
import requests
from bs4 import BeautifulSoup

from cpilatam import SETTINGS, logger
from cpilatam.names import Countries, CPIColumns
from cpilatam.parsers.base import BaseCPIParser
from cpilatam.schemas import CPI_SCHEMA


class PeruCPIParser(BaseCPIParser):
    BASE_URL = (
        "https://estadisticas.bcrp.gob.pe/estadisticas/"
        "series/mensuales/resultados/PN38705PM/html/{start_date}/{end_date}"
    )

    def __init__(
        self,
    ):
        start_date = date(1991, 1, 1).strftime("%Y-%-m")
        end_date = date.today().strftime("%Y-%-m")
        super().__init__(
            local_file_path=SETTINGS.PERU_LOCAL_PATH.as_posix(),
            url=self.BASE_URL.format(start_date=start_date, end_date=end_date),
            country=Countries.COLOMBIA.value,
        )

    def convert_spanish_date_to_numeric_date(self, date_str: str) -> str:
        """Converts a Spanish date string to a numeric date string.

        Args:
            date_str (str): A Spanish date string in the format "MMMYYYY".

        Returns:
            str: A numeric date string in the format "MM-YYYY".

        Example:
            >>> convert_spanish_date_to_numeric_date("Dic2021")
            "12-2021"
        """
        month = date_str[:3]
        year = date_str[3:]
        month_num = self.month_map[month]
        new_date = f"{month_num:02}-{year}"
        return new_date

    def parse_spanish_date_col(self, df: pd.DataFrame, col_name: str = "date"):
        df[col_name] = df[col_name].apply(self.convert_spanish_date_to_numeric_date)
        df[col_name] = pd.to_datetime(df[col_name], format="%m-%y")
        return df

    def set_reference_date(self) -> None:
        """Extracts the reference date from the column name of the CPI column.

        Example:
            >>> parser = PeruCPIParser()
            >>> parser.download()
            >>> parser.data.columns[1]
            Índice de precios Lima Metropolitana (índice Dic.2021 = 100) - Índice de Precios al Consumidor (IPC)'
            >>> parser.set_reference_date()
            >>> parser.reference_date
            "2021-12-01"
        """
        if self.data is not None:
            # extract the column that have the information when the data was obtained
            reference_date_str = self.data.columns[1]
            # define the pattern
            pattern = r"(Ene|Feb|Mar|Abr|May|Jun|Jul|Ago|Sep|Oct|Nov|Dic)\.(\d{4})"
            # match the pattern
            match = re.search(pattern, reference_date_str)
            # extract the reference date
            month_num = self.month_map[match.group(1)]
            year = match.group(2)
            # reference_date = f"{month_num:02d}-{year}"
            reference_date = pd.to_datetime(f"{year}-{month_num:02d}")
            # convert the reference date to a datetime object
            # reference_date = self.convert_spanish_date_to_numeric_date(reference_date_str)
            self.reference_date = reference_date
        else:
            logger.info("No data to parse. Please run the 'download' method first.")
            return None

    def download(self) -> None:
        """Downloads the data from the url and stores it in the self.data attribute."""
        logger.info("Downloading data from %s", self.url)
        # Send a GET request to the URL
        response = requests.get(self.url, timeout=10)

        # Check if the request was successful (status code 200)
        if response.status_code == 200:
            # Parse the HTML content
            soup = BeautifulSoup(response.text, "html.parser")

            # Find the table based on the specified XPath
            table = soup.select_one("#frmMensual > div.barra-resultados > table")

            # Check if the table is found
            if table:
                # Use pandas to read the HTML table into a DataFrame
                self.data = pd.read_html(str(table))[0]
            else:
                logger.error("Table not found on the webpage.")
        else:
            logger.error(
                "Failed to retrieve the webpage. Status code:",
                response.status_code,
            )

    def parse(self) -> None:
        """Parses the source cpi data into a pandas DataFrame with the universal schema."""
        logger.info(f"Parsing the data of {self.country}")
        if self.data is not None:
            # Extract reference date from the column name
            self.set_reference_date()

            # rename the columns
            self.data.columns = ["date", "CPI"]

            # Extract year and month from the "Fecha" column
            self.data = self.parse_spanish_date_col(self.data)

            # Add the reference date to the DataFrame
            self.data["reference_date"] = self.reference_date

            # Replace n.d. values with Nan
            self.data["CPI"] = self.data["CPI"].replace("n.d.", pd.NA)

            # Get the first and last nan value of CPI
            first_non_nan = self.data["CPI"].first_valid_index()
            last_non_nan = self.data["CPI"].last_valid_index()

            # Slice the dataframe
            self.data = self.data.loc[first_non_nan:last_non_nan]

            # Fill the NaN values with the previous value in the range (in place)
            self.data["CPI"].fillna(method="ffill", inplace=True)

            # Select and reorder columns
            self.data = self.data[["date", "reference_date", "CPI"]]

            # Rename columns according to the schema
            self.data = self.data.rename(
                columns={
                    "date": CPIColumns.DATE.value,
                    "reference_date": CPIColumns.REFERENCE_DATE.value,
                    "CPI": CPIColumns.CPI.value,
                }
            )

            # Validate and parse the DataFrame
            self.data = CPI_SCHEMA.validate(self.data)

        else:
            logger.info("No data to parse. Please run the 'download' method first.")
            return None


if __name__ == "__main__":
    # Example Usage:
    # Instantiate the class
    parser = PeruCPIParser()

    # Download data
    parser.download()

    # Parse and display the transformed data
    parser.parse()
    print(parser.data)



# -*- coding: utf-8 -*-
"""Top level package for recursiveseriation"""

import pandas as pd

from cpilatam.logger import configure_logging
from cpilatam.names import Countries, CPIColumns
from cpilatam.settings import init_settings

__app_name__ = "cpilatam"
__version__ = "2023.11.1"

SETTINGS = init_settings()
logger = configure_logging(__app_name__ + " - v" + __version__, SETTINGS, kidnap_loggers=True)

DF_CPI = {
    Countries.PERU.value: pd.read_csv(SETTINGS.PERU_LOCAL_PATH.as_posix()),
    Countries.COLOMBIA.value: pd.read_csv(SETTINGS.COLOMBIA_LOCAL_PATH.as_posix()),
}

for key, item in DF_CPI.items():
    if item[CPIColumns.DATE.value].max() <= pd.to_datetime("today").strftime("%Y-%m-%d"):
        logger.warn(f"The data is not up to date in the {key} country. Please run the update script.")


def update(countries: list = None):
    from cpilatam.parsers import __parsers__

    if countries is None:
        countries = DF_CPI.keys()
    for parser in __parsers__:
        if parser.country in countries:
            logger.info(f"Updating {parser.country} data...")
            parser.update()
            # replace the dataframe in the DF_CPI dictionary
            DF_CPI[parser.country] = parser.data



# -*- coding: utf-8 -*-
"""This module contains a parser for the Colombian CPI data."""

import re

import pandas as pd

from cpilatam import SETTINGS, logger
from cpilatam.names import Countries, CPIColumns
from cpilatam.parsers.base import BaseCPIParser


class ColombiaCPIParser(BaseCPIParser):
    def __init__(self):
        super().__init__(
            local_file_path=SETTINGS.COLOMBIA_LOCAL_PATH.as_posix(),
            url="https://www.dane.gov.co/files/operaciones/IPC/oct23/IPC_Indices.xlsx",
            country=Countries.COLOMBIA.value,
        )

    def download(self) -> None:
        """Downloads the data from the url and stores it in the self.data attribute."""
        logger.info("Downloading data from %s", self.url)
        # Read the Excel file into a pandas DataFrame
        self.data = pd.read_excel(self.url)

    def parse(self) -> None:
        """Parses the source cpi data into a pandas DataFrame with the universal schema."""
        logger.info(f"Parsing the data of {self.country}")
        if self.data is not None:
            # Get the date of obtanining the data
            data_obtain_date = self.data.iloc[6, 21]

            # Use regex to obtain day month year
            pattern = r"(\w+) de (\d{4})"
            matches = re.search(pattern, data_obtain_date)

            if matches:
                month, year = matches.groups()
                month = self.month_map[month]
                obtained_date = pd.to_datetime(f"{year}-{month:02d}")
            else:
                raise Exception("Date not found")

            # Get relevant data from the excel spreadsheet
            cleaned_data = self.data.iloc[7:20, 0:22]

            # Set the seventh row as column headers
            cleaned_data.columns = cleaned_data.iloc[0]

            # Remove the seventh row (it's now the header)
            cleaned_data = cleaned_data[1:]

            # Transform the "Mes" column according to the self.month_map dictionary of the abstract class
            cleaned_data["Mes"] = cleaned_data["Mes"].map(self.month_map)

            # Pivot the DataFrame to obtain a flattened time series
            flattened_data = cleaned_data.melt(id_vars=["Mes"], var_name="Año", value_name=CPIColumns.CPI.value)

            # Convert years to numeric
            flattened_data["Año"] = flattened_data["Año"].astype(int)

            def add_months(row):
                if pd.isnull(row["Año"]) or pd.isnull(row["Mes"]):
                    return pd.NaT  # Return NaT for missing values
                return pd.to_datetime(str(row["Año"]), format="%Y") + pd.DateOffset(months=row["Mes"] - 1)

            # Proceed to sum the year with the month to obtain a datetime object
            flattened_data[CPIColumns.DATE.value] = flattened_data.apply(add_months, axis=1)

            # Add the obtained date
            flattened_data[CPIColumns.REFERENCE_DATE.value] = obtained_date

            # Filter the data beyond the today date
            flattened_data = flattened_data[
                flattened_data[CPIColumns.DATE.value] <= pd.Timestamp.now().normalize()
            ]

            # Filter the non nan data
            first_non_nan = flattened_data[CPIColumns.CPI.value].first_valid_index()
            last_non_nan = flattened_data[CPIColumns.CPI.value].last_valid_index()

            # Slice the dataframe
            flattened_data = flattened_data.loc[first_non_nan:last_non_nan]

            # Fill the NaN values with the previous value in the range (in place)
            flattened_data[CPIColumns.CPI.value].fillna(method="ffill", inplace=True)

            # Updated self.data and slice
            self.data = flattened_data[
                [CPIColumns.DATE.value, CPIColumns.CPI.value, CPIColumns.REFERENCE_DATE.value]
            ]

        else:
            logger.info("No data to parse. Please run the 'download' method first.")
            return None


if __name__ == "__main__":
    parser = ColombiaCPIParser()
    # Download data
    parser.download()

    # Parse and display the transformed data
    parser.parse()
    print(parser.data)



# -*- coding: utf-8 -*-
"""Names for the CPI Latam package."""

from enum import Enum


class CPIColumns(Enum):
    """Enum for the CPI columns."""

    DATE = "date"
    CPI = "cpi"
    REFERENCE_DATE = "reference_date"


class Countries(Enum):
    """Enum for the Countries columns."""

    PERU = "peru"
    COLOMBIA = "colombia"



# -*- coding: utf-8 -*-
"""Logger configuration."""

import logging.config
import uuid
from functools import wraps

import structlog

from cpilatam.exc import EnvVarNotFound
from cpilatam.settings import LogDest, LogFormatter, Settings

try:
    import colorama  # noqa

    COLORAMA_INSTALLED = True

except ImportError:
    COLORAMA_INSTALLED = False


COMMON_CHAIN = [
    structlog.threadlocal.merge_threadlocal_context,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    structlog.processors.TimeStamper(fmt="iso"),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
]


def _control_logging(
    settings: Settings,
):
    level = settings.LOG_LEVEL
    formatter = settings.LOG_FORMAT
    dest = settings.LOG_DESTINATION
    log_file = settings.LOG_PATH

    if formatter == LogFormatter.JSON.value:
        fmt = {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.processors.JSONRenderer(),
            "foreign_pre_chain": COMMON_CHAIN,
        }
    elif formatter == LogFormatter.COLOR.value:
        fmt = {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.dev.ConsoleRenderer(colors=COLORAMA_INSTALLED),
            "foreign_pre_chain": COMMON_CHAIN,
        }

    else:
        raise NotImplementedError("Pydantic should not allow this.")  # pragma: no cover

    if dest == LogDest.CONSOLE.value:
        hndler = {
            "level": level,
            "class": "logging.StreamHandler",
            "formatter": "default",
        }
    elif dest == LogDest.FILE.value:
        if not log_file:
            raise EnvVarNotFound(env_var=f"{settings.Config.env_prefix}LOG_PATH")

        hndler = {
            "level": level,
            "class": "logging.handlers.RotatingFileHandler",
            "filename": str(log_file),
            "formatter": "default",
            "maxBytes": 10e6,
            "backupCount": 100,
        }
        log_file.parent.mkdir(parents=True, exist_ok=True)
    else:
        raise NotImplementedError("Pydantic should not allow this.")  # pragma: no cover

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {"default": fmt},
            "handlers": {"default": hndler},
            "loggers": {
                "": {
                    "handlers": ["default"],
                    "level": level,
                    "propagate": True,
                }
            },
        }
    )
    structlog.configure_once(
        processors=[
            structlog.stdlib.filter_by_level,
            *COMMON_CHAIN,
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )


def configure_logging(
    name: str,
    settings: Settings,
    kidnap_loggers: bool = False,
):
    """logger configuration.

    :param name: logger name.
    :param settings: Setting class with logger information
            with configuration information.

    """

    if kidnap_loggers:
        _control_logging(settings)

    logger = structlog.get_logger(name)
    logger.trace = trace_using(logger)  # pylint: disable=assignment-from-none
    return logger


def trace_using(logger):
    """Factory of decorators to trace callables."""

    def real_decorator(func):  # pylint: disable=unused-variable
        """Decorate a callable for the args, kwargs and returns."""

        @wraps(func)
        def wrapper(*args, **kwargs):
            uuid_ = str(uuid.uuid4())
            qual = func.__qualname__
            args_repr = ",".join(repr(a) for a in args)
            kwargs_repr = ",".join(k + "=" + repr(v) for k, v in kwargs.items())
            repr_ = f"{qual}({args_repr},{kwargs_repr})"
            with structlog.threadlocal.tmp_bind(
                logger,
                repr=repr_,
                uuid=uuid_,
                func=qual,
                args=args,
                kwargs=kwargs,
            ) as tmp_log:
                tmp_log.info(event="CALLED")
                retval = func(*args, **kwargs)
                tmp_log.info(event="RETURN", value=retval)
            return retval

        return wrapper

    return real_decorator



# -*- coding: utf-8 -*-
"""Project settings."""

import logging
from enum import Enum, IntEnum
from pathlib import Path
from typing import Optional

from dotenv import find_dotenv, load_dotenv
from pydantic import BaseSettings


def init_dotenv():
    """Loc n' load dotenv file.

    Sets the location for a dotenv file containig envvars loads its
    contents.

    Returns:
        Location of the dotenv file.

    """

    candidate = find_dotenv(usecwd=True)

    if not candidate:
        print(".env file not found, env vars must be seted manually")
        return

    load_dotenv(candidate)


class LogLevel(IntEnum):
    """Explicitly define allowed logging levels."""

    CRITICAL = logging.CRITICAL

    ERROR = logging.ERROR

    WARNING = logging.WARNING

    INFO = logging.INFO

    DEBUG = logging.DEBUG

    TRACE = 1 + logging.NOTSET

    NOTSET = logging.NOTSET


class LogDest(Enum):
    """Define allowed destinations for logs."""

    CONSOLE = "CONSOLE"
    """Log to console"""

    FILE = "FILE"
    """Log to file"""


class LogFormatter(Enum):
    """Define allowed destinations for logs."""

    JSON = "JSON"
    """JSONs, eg for filebeat or similar, for machines."""

    COLOR = "COLOR"
    """pprinted, colored, for humans"""


class Settings(BaseSettings):
    """Project settings variables."""

    PACKAGE_PATH = Path(__file__).parent
    """Package path (python files)."""

    PROJECT_PATH = PACKAGE_PATH.parent
    """Project path (all files)."""

    LOG_PATH: Optional[Path]
    """Path to logfile, only works if ``LOG_DESTINATION=FILE``."""

    LOG_FORMAT: LogFormatter = LogFormatter.JSON.value
    """Log style."""

    LOG_LEVEL: LogLevel = LogLevel.INFO.value
    """Log level from ``logging`` module."""

    LOG_DESTINATION: LogDest = LogDest.CONSOLE.value
    """Destination for logs."""

    COLOMBIA_LOCAL_PATH: Path = Path(PACKAGE_PATH, "data", "colombia.csv")
    """Path to local file with Colombia CPI data."""

    PERU_LOCAL_PATH: Path = Path(PACKAGE_PATH, "data", "peru.csv")
    """Path to local file with Peru CPI data."""

    class Config:
        """Inner configuration."""

        env_prefix = "CPILATAM_"
        use_enum_values = True


def init_settings() -> Settings:
    """Initilize project settings"""
    init_dotenv()
    return Settings()



# -*- coding: utf-8 -*-
"""This module contains the base class for CPI parsers."""

from abc import ABC, abstractmethod
from datetime import date

import pandas as pd
from pandera.typing import DataFrame

from cpilatam.schemas import CPI_SCHEMA


class BaseCPIParser(ABC):
    """Base class for CPI parsers."""

    month_map = {
        # Peru format
        "Ene": 1,
        "Feb": 2,
        "Mar": 3,
        "Abr": 4,
        "May": 5,
        "Jun": 6,
        "Jul": 7,
        "Ago": 8,
        "Sep": 9,
        "Oct": 10,
        "Nov": 11,
        "Dic": 12,
        # Colombia/Chile format
        "Enero": 1,
        "Febrero": 2,
        "Marzo": 3,
        "Abril": 4,
        "Mayo": 5,
        "Junio": 6,
        "Julio": 7,
        "Agosto": 8,
        "Septiembre": 9,
        "Octubre": 10,
        "Noviembre": 11,
        "Diciembre": 12,
    }

    def __init__(self, local_file_path: str, url: str, country: str):
        """Initializes the parser.

        Args:
            local_file_path (str): The path to the local file.
            url (str): The url to the source data.
            country (str): The country of the CPI data.

        Attributes:
            local_file_path (str): The path to the local file.
            url (str): The url to the source data.
            data (pd.DataFrame): The data in a pandas DataFrame with the universal schema.
            reference_date (date): The reference/pivot for the CPI values.
            country (str): The country of the CPI data.
        """
        self.local_file_path: str = local_file_path
        self.url: str = url
        self.data: pd.DataFrame = None
        self.reference_date: date = None
        self.country: str = country

    @abstractmethod
    def parse(self) -> None:
        """Parses the source cpi data into a pandas DataFrame with the universal schema.

        Returns:
            pd.DataFrame: A pandas DataFrame with the universal schema.
        """
        pass

    @abstractmethod
    def download(self) -> None:
        """Downloads the raw data from the internet and saves it to a local file in csv format."""
        pass

    def save(self) -> None:
        """Saves the parsed data to a local csv file."""
        self.data.to_csv(self.local_file_path, index=False)

    def update(self) -> None:
        """Updates the data by downloading the raw data and reading it into a pandas DataFrame."""
        self.download()
        self.parse()
        self.save()

    def get_data(self) -> DataFrame[CPI_SCHEMA]:
        """Returns the data in a pandas DataFrame with the universal schema.

        Returns:
            pd.DataFrame: A pandas DataFrame with the universal schema.
        """
        return self.data



# -*- coding: utf-8 -*-
"""This module contains the schema for the CPI data."""

from pandera import Check, Column, DataFrameSchema, DateTime, Float

from cpilatam.names import CPIColumns

CPI_SCHEMA = DataFrameSchema(
    columns={
        CPIColumns.DATE.value: Column(
            DateTime,
            nullable=False,
            checks=[
                Check(lambda x: x.dt.day == 1, error="The day must be the first day of the month."),
            ],
        ),
        CPIColumns.CPI.value: Column(Float, nullable=False),
        CPIColumns.REFERENCE_DATE.value: Column(
            DateTime,
            nullable=False,
            checks=[
                Check(lambda x: x.dt.day == 1, error="The day must be the first day of the month."),
            ],
        ),
    },
    coerce=True,
    strict=True,
)
